/**************************************

  XIScheckEventNo

    ver 0.0 2005/08/28	K. Hayashida

    ver 0.1 2005/12/16	K. Hayashida
	Check event no to search for telemetry saturation
    ver 0.2 2005/12/18	K. Hayashida
	Implement segment selection
    ver 0.3 2007/01/31	Y. ISHISAKI
	add gapsec parameter, remove gtifile parameter
	BnkGet XIS:FRAMES to determine required memory in _init()
	BnkGet XIS:OUTFITS:FILENAME:PTR to determine output GTI file name
	add functions, checkTimeJump(), checkSaturation()
	add functions, showTimeJump(), showFrameInfo(), addToStatistics()
	consider last frame in _exit()
    ver 1.0 2007/01/31	Y. ISHISAKI
	write FITS GTI file
	add clobber parameter
    ver 1.1 2007/04/30	Y. ISHISAKI
	print comment
    ver 1.2 2007/05/07	Y. ISHISAKI
	bug fix in handling frame time jump in _exit()
	static declaration of gti
    ver 1.3 2009/03/08	Y. ISHISAKI
	update NAXIS2 keywords in GTI extension in write_gti()
	free allocated memory in _exit()
**************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "cli.h"
#include "com.h"
#include "bnk.h"
#include "evs.h"
#include "anl.h"
#include "fitsio.h"
#include "pil.h"
#include "aste_gti.h"
#include "xisTelemFormat.h"
#include "aeFitsHeaderUtil.h"

static char pname[] = "XIScheckEventNo";
char XIScheckEventNo_version[] = "version 1.3";

static int clobber;
static double gapsec = 0.1;
static int check_seg[XIStotalSegNo];

static fitsfile *ofp;
static char *infile;
static char *outfile;
static int col_start, col_stop;
static GTI_DATA gti;

static double *frame_st_aetime;
static double *frame_end_aetime;
static int *frame_event_totno;
static int *event_totno_inlist;
static int initflag;
static int ithframe;
static long frames;
static int n_tested = 0;
static int n_passed = 0;
static int n_time_jump = 0;
static int n_saturated = 0;
static double t_tested = 0.0;
static double t_passed = 0.0;
static double t_time_jump = 0.0;
static double t_saturated = 0.0;

static struct {
  char *N_FRAMES;
  char *N_TESTED;
  char *N_PASSED;
  char *N_T_JUMP;
  char *N_SATURA;
  char *T_TESTED;
  char *T_PASSED;
  char *T_T_JUMP;
  char *T_SATURA;
} cm = {
  "number of frames in the input event file",
  "number of non-zero frames tested",
  "number of frames passed the test",
  "number of frames detected time jump",
  "number of frames telemetry saturated",
  "exposure of non-zero frames tested",
  "exposure of frames passed the test",
  "loss of exposure due to time jump",
  "exposure of frames telemetry saturated"
};

static void
showParam(void)
{
  printf ("\nANL:  *** %s show parameter ***\n\n", pname);
  printf("%20s   %.6f (s)\n", "GAPSEC", gapsec);
  printf("%20s   %s\n", "SEGMENT_A", check_seg[0] ? "YES" : "NO");
  printf("%20s   %s\n", "SEGMENT_B", check_seg[1] ? "YES" : "NO");
  printf("%20s   %s\n", "SEGMENT_C", check_seg[2] ? "YES" : "NO");
  printf("%20s   %s\n", "SEGMENT_D", check_seg[3] ? "YES" : "NO");
  printf("%20s   %s\n", "CLOBBER", clobber ? "YES" : "NO");
  printf("\n");
}

static int
checkTimeJump(int j)
{
  return ( 0 < j && frame_end_aetime[j-1] + gapsec < frame_st_aetime[j] );
}

static int
checkSaturation(int j)
{
  return (event_totno_inlist[j] != frame_event_totno[j]);
}

static void
showTimeJump(int j)
{
  anl_msg_info("\
frame time jump, t=%.3f - %.3f by %.3f s\n",
	frame_end_aetime[j-1], frame_st_aetime[j],
	frame_st_aetime[j] - frame_end_aetime[j-1]);
}

static void
showFrameInfo(int j)
{
  if ( checkSaturation(j) ) {
    anl_msg_info("\
saturated frame, t=%.3f - %.3f  %d (%d/%d) seg=%d%d%d%d\n",
	frame_st_aetime[j], frame_end_aetime[j],
	frame_event_totno[j] - event_totno_inlist[j],
	event_totno_inlist[j], frame_event_totno[j],
	check_seg[0], check_seg[1], check_seg[2], check_seg[3]);
  } else {
    anl_msg_debug("\
%.3f - %.3f  %d (%d/%d) seg=%d%d%d%d\n",
	frame_st_aetime[j], frame_end_aetime[j],
	frame_event_totno[j] - event_totno_inlist[j],
	event_totno_inlist[j], frame_event_totno[j],
	check_seg[0], check_seg[1], check_seg[2], check_seg[3]);
  }
}

static void
addToStatistics(int j)
{
  double frame_exposure = 0.0;

  if ( 0.0 != frame_end_aetime[j] && 0.0 != frame_st_aetime[j] ) {
    frame_exposure = frame_end_aetime[j] - frame_st_aetime[j];
  }

  n_tested++;
  t_tested += frame_exposure;
  if ( checkSaturation(j) ) {
    n_saturated++;
    t_saturated += frame_exposure;
  } else {
    n_passed++;
    t_passed += frame_exposure;
  }

  if ( checkTimeJump(j) ) {
    n_time_jump++;
    if ( 0.0 != frame_st_aetime[j] && 0.0 != frame_end_aetime[j-1] ) {
      t_time_jump += frame_st_aetime[j] - frame_end_aetime[j-1];
    }
    showTimeJump(j);
  }

  showFrameInfo(j);
}

static int
write_gti(fitsfile *ofp)
{
  char buf[PIL_LINESIZE];
  char *k, tdisp1[FLEN_KEYWORD], tdisp2[FLEN_KEYWORD];
  int istat = 0;

  istat = aefits_write_module_history(ofp, pname);
  if ( istat ) goto quit;
  sprintf(buf, "  infile='%s'", infile);
  fits_write_history(ofp, buf, &istat);
  sprintf(buf, "  outfile='%s'", outfile);
  fits_write_history(ofp, buf, &istat);
  sprintf(buf, "  gapsec=%.6f  clobber=%s", gapsec, clobber ? "yes" : "no");
  fits_write_history(ofp, buf, &istat);
  sprintf(buf, "  segment_a=%s  segment_b=%s  segment_c=%s  segment_d=%s",
	  check_seg[0] ? "yes" : "no",
	  check_seg[1] ? "yes" : "no",
	  check_seg[2] ? "yes" : "no",
	  check_seg[3] ? "yes" : "no");
  fits_write_history(ofp, buf, &istat);

  sprintf(tdisp1, "TDISP%d", col_start);
  sprintf(tdisp2, "TDISP%d", col_stop);
  if (
fits_update_key_str(ofp, k=tdisp1, "F16.6", "display format of START",&istat)||
fits_update_key_str(ofp, k=tdisp2, "F16.6", "display format of STOP", &istat)||
fits_update_key_fixdbl(ofp, k="TSTART", gti.tstart, 6, NULL, &istat) ||
fits_update_key_fixdbl(ofp, k="TSTOP", gti.tstop, 6, NULL, &istat) ||
fits_update_key_fixdbl(ofp, k="TELAPSE", gti.telapse, 6, NULL, &istat) ||
fits_update_key_fixdbl(ofp, k="ONTIME", gti.ontime, 6, NULL, &istat) ||
fits_update_key_lng(ofp, k="NAXIS2", gti.ngti, NULL, &istat) ||
fits_write_key_lng(ofp, k="N_FRAMES", frames, cm.N_FRAMES, &istat) ||
fits_write_key_lng(ofp, k="N_TESTED", n_tested, cm.N_TESTED, &istat) ||
fits_write_key_lng(ofp, k="N_PASSED", n_passed, cm.N_PASSED, &istat) ||
fits_write_key_lng(ofp, k="N_T_JUMP", n_time_jump, cm.N_T_JUMP, &istat) ||
fits_write_key_lng(ofp, k="N_SATURA", n_saturated, cm.N_SATURA, &istat) ||
fits_write_key_fixdbl(ofp, k="T_TESTED", t_tested, 6, cm.T_TESTED, &istat) ||
fits_write_key_fixdbl(ofp, k="T_PASSED", t_passed, 6, cm.T_PASSED, &istat) ||
fits_write_key_fixdbl(ofp, k="T_T_JUMP", t_time_jump, 6, cm.T_T_JUMP, &istat)||
fits_write_key_fixdbl(ofp, k="T_SATURA", t_saturated, 6, cm.T_SATURA, &istat)||
      0 ) {
    anl_msg_error("\
%s: fits_write_key('%s') failed (%d)\n", pname, k, istat);
    goto quit;
  }
  fits_write_col_dbl(ofp, col_start, 1, 1, gti.ngti, gti.start, &istat);
  if ( istat ) {
    anl_msg_error("\
%s: fits_write_col('START') failed (%d)\n", pname, istat);
    goto quit;
  }
  fits_write_col_dbl(ofp, col_stop, 1, 1, gti.ngti, gti.stop, &istat);
  if ( istat ) {
    anl_msg_error("\
%s: fits_write_col('STOP') failed (%d)\n", pname, istat);
    goto quit;
  }
  fits_write_chksum(ofp, &istat);
  if ( istat ) {
    anl_msg_error("\
%s: fits_write_chksum() failed for GTI extension (%d)\n", pname, istat);
    goto quit;
  }

 quit:
  return istat;
}

void
XIScheckEventNo_startup(int *status)
{
  int iseg;

  for (iseg = 0; iseg < XIStotalSegNo; iseg++) {
    check_seg[iseg] = 1;
  }

  *status = ANL_OK;
}

void
XIScheckEventNo_com(int *status)
{
  static char *keytbl[] = {
    "SHOW",
    "GAPSEC",
    "SEGMENT_A",
    "SEGMENT_B",
    "SEGMENT_C",
    "SEGMENT_D",
    "EXIT"
  };
  static char *help[] = {
    "Show current setting",
    "Allowed gap between frames in second",
    "Check Segment A",
    "Check Segment B",
    "Check Segment C",
    "Check Segment D",
    "Exit from this menu"
  };
  static int nkey = sizeof(keytbl) / sizeof(keytbl[0]);

  if ( *status ) {	/* ftools */
    char *k;
    if (
PILGetReal(k="gapsec", &gapsec) ||
PILGetBool(k="segment_a", &check_seg[0]) ||
PILGetBool(k="segment_b", &check_seg[1]) ||
PILGetBool(k="segment_c", &check_seg[2]) ||
PILGetBool(k="segment_d", &check_seg[3]) ||
PILGetBool(k="clobber", &clobber) ||
	 0 ) {
      anl_msg_error("\
%s: PILGet('%s') failed\n", pname, k);
      *status = ANL_QUIT;
      return;
    }
    *status = ANL_OK;
    return;
  }

  for(;;) {
    char *key;
    int ans[2];
    CMinquir(pname, nkey, keytbl, help, 1, ans);
    key = keytbl[ans[1]-1];
    if ( 0 == strcmp("GAPSEC", key) ) {
      CLfdprd(key, &gapsec);
    } else if ( 0 == strcmp("SEGMENT_A", key) ) {
      CLlogrd(key, &check_seg[0]);
    } else if ( 0 == strcmp("SEGMENT_B", key) ) {
      CLlogrd(key, &check_seg[1]);
    } else if ( 0 == strcmp("SEGMENT_C", key) ) {
      CLlogrd(key, &check_seg[2]);
    } else if ( 0 == strcmp("SEGMENT_D", key) ) {
      CLlogrd(key, &check_seg[3]);
    } else if ( 0 == strcmp ("SHOW", key) ) {
      showParam();
    } else if ( 0 == strcmp("EXIT", key) ) {
      break;
    }
  }

  *status = ANL_OK;
}

void
XIScheckEventNo_init(int *status)
{
  long i;
  int used;

  EvsDef("XIScheckEventNo:BEGIN");
  EvsDef("XIScheckEventNo:ENTRY");
  EvsDef("XIScheckEventNo:OK");

  showParam();

  BnkGet("XIS:FRAMES", sizeof(frames), &used, &frames);
  if ( 0 == frames ) {	/* NO FRAMES EXTENSION */
    anl_msg_error("\
%s: no FRAMES extension found.\n", pname);
    *status = ANL_QUIT;
    return;
  }

  /* Memory Allocation */
  frame_st_aetime = malloc(2 * frames * sizeof(*frame_st_aetime));
  frame_end_aetime = &frame_st_aetime[frames];
  event_totno_inlist = malloc(2 * frames * sizeof(*event_totno_inlist));
  frame_event_totno = &event_totno_inlist[frames];
  gti.start = malloc(2 * frames * sizeof(*gti.start));
  gti.stop = &gti.start[frames];

  if ( NULL == frame_st_aetime ||
       NULL == event_totno_inlist ||
       NULL == gti.start ) {
    anl_msg_error("\
%s: memory allocation error", pname);
    *status = ANL_QUIT;
    return;
  }

  for (i = 0; i < frames; i++) {
    frame_st_aetime[i] = 0.0;
    frame_end_aetime[i] = 0.0;
    event_totno_inlist[i] = 0;
    frame_event_totno[i] = 0;
  }

  BnkGet("XIS:FILENAME:PTR", sizeof(infile), &used, &infile);
  BnkGet("XIS:OUTFITS:FILENAME:PTR", sizeof(outfile), &used, &outfile);

  if ( 0 == CLstricmp(outfile, "NONE") ) {
    ofp = NULL;
  } else {
    fitsfile *ifp;
    char *k;
    int hdutype;
    int extver = 0;
    int istat = 0;

    BnkGet("XIS:FITS:PTR", sizeof(ifp), &used, &ifp);
    fits_reopen_file(ifp, &ifp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_reopen_file() failed (%d)\n", pname, istat);
      goto quit;
    }
    fits_movabs_hdu(ifp, 1, &hdutype, &istat);	/* move to primary */
    if ( istat ) {
      anl_msg_error("\
%s: fits_movabs_hdu(hdunum=1) failed (%d)\n", pname, istat);
      goto quit;
    }

    if ( clobber ) {
      unlink(outfile);
    }
    fits_create_file(&ofp, outfile, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_create_file('%s') failed (%d)\n", pname, outfile, istat);
      goto quit;
    }

    fits_copy_header(ifp, ofp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_copy_header() failed for primary extension (%d)\n", pname, istat);
      goto quit;
    }
    fits_write_date(ofp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_write_date() failed for primary extension (%d)\n", pname, istat);
      goto quit;
    }
    fits_write_chksum(ofp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_write_chksum() failed for primary extension (%d)\n", pname, istat);
      goto quit;
    }

    fits_movnam_hdu(ifp, BINARY_TBL, "GTI", extver, &istat);
    if ( istat ) {
      istat = 0;
      fits_movnam_hdu(ifp, BINARY_TBL, "STDGTI", extver, &istat);
    }
    if ( istat ) {
      anl_msg_error("\
%s: fits_movnam_hdu('GTI') failed (%d)\n", pname, istat);
      goto quit;
    }

    fits_copy_header(ifp, ofp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_copy_header() failed for GTI extension (%d)\n", pname, istat);
      goto quit;
    }

    fits_close_file(ifp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_close_file() failed for input event file (%d)\n", pname, istat);
      goto quit;
    }

    if ( fits_get_colnum(ofp, CASESEN, k="START", &col_start, &istat) ||
	 fits_get_colnum(ofp, CASESEN, k="STOP", &col_stop, &istat) ) {
      anl_msg_error("\
%s: fits_get_colnum('%s') failed (%d)\n", pname, k, istat);
      goto quit;
    }

  }

  initflag = ANL_TRUE;

  *status = ANL_OK;
  return;

 quit:
  *status = ANL_QUIT;
  return;
}

void
XIScheckEventNo_his(int *status)
{
  *status = ANL_OK;
}

void
XIScheckEventNo_bgnrun(int *status)
{
  *status = ANL_OK;
}

void
XIScheckEventNo_ana(int nevent, int eventid, int *status)
{
  int iseg, used;
  int ev_tot_no[XIStotalSegNo];
  double t0, t1;

  EvsfSetM("XIScheckEventNo:ENTRY");

  BnkfGetM("XIS:FRAMES:FRAME_ST_AETIME", sizeof(double), &used, &t0);
  /*printf("frame st aetime %f\n", t0); */

  if ( initflag ) {
    ithframe = 0;
    initflag = ANL_FALSE;
  } else if ( frame_st_aetime[ithframe] + gapsec < t0 ) {
    addToStatistics(ithframe);
    ithframe++;
  } else {
    goto skip;
  }

  if ( frames <= ithframe ) {
    anl_msg_error("\
%s: too many frames (%d), something is wrong\n", pname, ithframe);
    anl_put_exit_status(1);
    *status = ANL_QUIT;
    return;
  }
  /*=========== initial frame or 1st event in a frame ===== */
  BnkfGetM("XIS:FRAMES:FRAME_END_AETIME", sizeof(double), &used, &t1);
  /* printf("frame end aetime %lf\n", t1); */
  BnkfGetM("XIS:FRAMES:EVENT_TOT_NO", sizeof(ev_tot_no), &used, &ev_tot_no);
  /* printf("event tot no %d %d %d %d\n",
     ev_tot_no[0], ev_tot_no[1], ev_tot_no[2], ev_tot_no[3]); */
  frame_st_aetime[ithframe] = t0;
  frame_end_aetime[ithframe] = t1;
  frame_event_totno[ithframe] = 0;
  for (iseg = 0; iseg < XIStotalSegNo; iseg++) {
    if ( check_seg[iseg] ) {
      frame_event_totno[ithframe] += ev_tot_no[iseg];
    }
  }
  event_totno_inlist[ithframe] = 0;

 skip:

  BnkfGetM("XIS:SEGMENT", sizeof(int), &used, &iseg);
  if ( check_seg[iseg] ) {
    event_totno_inlist[ithframe]++;
  }

  EvsfSetM("XIScheckEventNo:OK");

  *status = ANL_OK;
}

void
XIScheckEventNo_endrun(int *status)
{
  *status = ANL_OK;
}

void
XIScheckEventNo_exit(int *status)
{
  int j;
  int igti = 0;
  int flag_goodtime = ANL_FALSE;
  int istat = 0;

  addToStatistics(ithframe);	/* add last frame */

  for (j = 0; j < n_tested && checkSaturation(j); j++) {
    ;
  }

  if ( j < n_tested ) {
    gti.start[igti] = frame_st_aetime[j];
    flag_goodtime = ANL_TRUE;
    j++;
  }

  while ( j < n_tested ) {

    if ( flag_goodtime ) {	/* during good time */

      if ( checkSaturation(j) ) {	/* staurated frame, not usable */
	flag_goodtime = ANL_FALSE;
	gti.stop[igti] = frame_end_aetime[j-1];
	igti++;

      } else {				/* not saturated, usable frame */

	if ( checkTimeJump(j) ) {	/* frame time jump detected */
	  gti.stop[igti] = frame_end_aetime[j-1];
	  igti++;
	  gti.start[igti] = frame_st_aetime[j];
	}

      }

    } else {			/* during bad time */

      if ( 0 == checkSaturation(j) ) {	/* not saturated, usable frame */
	flag_goodtime = ANL_TRUE;
	gti.start[igti] = frame_st_aetime[j];
      }
    }

    j++;
  }

  if ( flag_goodtime ) {
    gti.stop[igti] = frame_end_aetime[n_tested-1];
    igti++;
  }

  gti.ngti = igti;
  gti.tstart = gti.start[0];
  gti.tstop = gti.stop[igti-1];
  gti.telapse = gti.tstop - gti.tstart;
  gti.ontime = t_passed;

  if ( NULL != ofp ) {
    if ( write_gti(ofp) ) {
      goto quit;
    }

    fits_close_file(ofp, &istat);
    if ( istat ) {
      anl_msg_error("\
%s: fits_close_file('%s') failed (%d)\n", pname, outfile, istat);
      goto quit;
    }

    anl_msg_info("\
%s: GTI file '%s' created\n", pname, outfile);
  }

  anl_msg_always("\
N_FRAMES = %16ld / %s\n\
N_TESTED = %16d / %s\n\
N_PASSED = %16d / %s\n\
N_T_JUMP = %16d / %s\n\
N_SATURA = %16d / %s\n\
T_TESTED = %16.6f / %s\n\
T_PASSED = %16.6f / %s\n\
T_T_JUMP = %16.6f / %s\n\
T_SATURA = %16.6f / %s\n",
	frames, cm.N_FRAMES,
	n_tested, cm.N_TESTED,
	n_passed, cm.N_PASSED,
	n_time_jump, cm.N_T_JUMP,
	n_saturated, cm.N_SATURA,
	t_tested, cm.T_TESTED,
	t_passed, cm.T_PASSED,
	t_time_jump, cm.T_T_JUMP,
	t_saturated, cm.T_SATURA);

  free(gti.start);
  free(event_totno_inlist);
  free(frame_st_aetime);

  *status = ANL_OK;
  return;

 quit:
  *status = ANL_QUIT;
  return;
}

/*	for Emacs
;;; Local Variables: ***
;;; mode:C ***
;;; tab-width:8 ***
;;; c-indent-level:2  ***
;;; c-basic-offset:2  ***
;;; End: ***
*/
